// Copyright (C) 2009 Playedit.

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "config.h"

#include "app_window.h"
#include "playlist.h"
#include "music.h"
#include "basic_dialogs.h"

#include <glibmm/i18n.h>
#include <glibmm/miscutils.h>
#include <glibmm/threads.h>
#include <glibmm/timer.h>
#include <gtkmm/action.h>
#include <gtkmm/main.h>
#include <gtkmm/icontheme.h>
#include <iostream>

#define ICON_DIR DATADIR "/pixmaps"
#define PLAYEDIT_ICON_FILE "playedit.svg"
#define RENAME_DEVICE_ICON_FILE "rename-device.svg"

namespace Playedit
{

App::App()
	: m_Device(App::PulseProgress, this),
	m_Box(Gtk::ORIENTATION_VERTICAL),
	m_Paned(Gtk::ORIENTATION_HORIZONTAL),
	m_MusicBox(Gtk::ORIENTATION_HORIZONTAL, 1),
	m_PlaylistBox(Gtk::ORIENTATION_HORIZONTAL, 1)
{
	set_default_size(900, 700);
	set_has_resize_grip();

	// Set the default window icon
	try
	{
		set_default_icon_from_file(Glib::build_filename(ICON_DIR, PLAYEDIT_ICON_FILE));
	}
	catch (const Glib::Error& err)
	{
		std::cerr << _("Unable to set window icon: ") << err.what() << std::endl;
	}

	// Add the rename device icon
	Glib::RefPtr<Gtk::IconTheme> icon_theme = Gtk::IconTheme::get_default();
	try
	{
		Glib::RefPtr<Gdk::Pixbuf> rename_device_icon =
			Gdk::Pixbuf::create_from_file(Glib::build_filename(ICON_DIR, RENAME_DEVICE_ICON_FILE), 64, 64);
		icon_theme->add_builtin_icon("RenameDevice", 64, rename_device_icon);
	}
	catch (const Glib::Error& err)
	{
		std::cerr << _("Unable to load theme icon: ") << err.what() << std::endl;
	}

	// Set the initial application title
	set_title("PlayEdit");

	// Add the vbox that will contain everything
	add(m_Box);

	// Add the status bar
	m_Box.pack_end(m_Statusbar, Gtk::PACK_SHRINK);

	// Set up the ui manager
	CreateUIManager();
	add_accel_group(m_UIManager->get_accel_group());

	// Set up the menu and toolbar
	m_Box.pack_start(*m_UIManager->get_widget("/MainMenu"), Gtk::PACK_SHRINK);
	m_Box.pack_start(*m_UIManager->get_widget("/MainToolbar"), Gtk::PACK_SHRINK);

	// Create the progress bar
	m_Statusbar.pack_end(m_Progress, Gtk::PACK_SHRINK, 1);

	// Add an hpaned for the music and playlists
	m_Paned.set_position(450);
	m_Box.pack_start(m_Paned);

	// Add an hbox to contain the music view and its toolbar
	m_Paned.add1(m_MusicBox);
	
	// Setup the Music Tree View
	m_MusicView = new MusicView(*this);
	m_MusicViewport.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	m_MusicViewport.add(*Gtk::manage(m_MusicView));
	m_MusicBox.pack_end(m_MusicViewport);

	// Setup the music toolbar
	m_MusicBox.pack_start(*GetMusicToolbar(), Gtk::PACK_SHRINK);

	// Add an hbox to contain the playlist and its toolbar
	m_Paned.add2(m_PlaylistBox);

	// Setup the Playlist Tree View
	m_PlaylistView = new PlaylistView(*this);
	m_PlaylistViewport.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC);
	m_PlaylistViewport.add(*Gtk::manage(m_PlaylistView));
	m_PlaylistBox.pack_start(m_PlaylistViewport);

	// Setup the playlist toolbar
	m_PlaylistBox.pack_end(*GetPlaylistToolbar(), Gtk::PACK_SHRINK);

	// Display Main Window
	show_all();

	// Make the device interactions insensitive
	m_DeviceInteractionGroup->set_sensitive(false);

	// Disable device actions
	m_DeviceActionGroup->set_sensitive(false);
	m_DeviceActionGroup->set_visible(false);

	// Set up the udev client
	const char* const subsystems[] = { "usb", NULL };
	m_GUDevClient = g_udev_client_new(subsystems);
	g_signal_connect(m_GUDevClient, "uevent", G_CALLBACK (UEventCB), this);

	// Scan for mtp devices on first run
	GList* list = g_udev_client_query_by_subsystem(m_GUDevClient, "usb");
	for (GList* it = list; it != NULL; it = g_list_next(it))
	{
		// Pretend the device was just added
		UEventCB(m_GUDevClient, "add", G_UDEV_DEVICE(it->data), this);
		g_object_unref(it->data);
	}
	g_list_free(list);
}

App::~App()
{
	// Unref the udev client
	g_object_unref(m_GUDevClient);
}

void App::ProcessPending()
{
	while (Gtk::Main::events_pending())
		Gtk::Main::iteration();
}

void App::SetStatus(const Glib::ustring& message)
{
	m_Statusbar.pop();
	m_Statusbar.push(message);
	ProcessPending();
}

void App::UnsetStatus()
{
	m_Statusbar.pop();
}

// When udev detects a new device check to see if it is an mtp device.  If
// there is not already a connected device connect to it.  When udev detects
// that the device is removed disconnect.
void App::UEventCB(GUdevClient* /* client */, const char* action, GUdevDevice* device,
		Playedit::App* app_window)
{
	GUdevDeviceNumber device_number = g_udev_device_get_device_number(device);
	if (g_strcmp0(g_udev_device_get_subsystem(device), "usb") != 0 || !device_number)
		return;

	const char* devnum_str = g_udev_device_get_property(device, "DEVNUM");
	if (devnum_str == NULL)
		return;
	int devnum = strtol(devnum_str, NULL, 10);

	if (g_str_equal(action, "add") && !app_window->m_Device)
		app_window->ConnectDevice(devnum);
	else if (g_str_equal(action, "remove") && app_window->m_Device &&
			app_window->m_Device.DevNum() == devnum)
		app_window->DisconnectDevice();
}

void App::QuitPlayedit()
{
	hide();
}

// Force the program to close while the device is connecting.
static bool ForceQuit(GdkEventAny* /* event */, bool& initing)
{
	initing = !ConfirmationDialog(_("Force Quit"),
			_("Force the program to close?\n"
				"WARNING: This may cause damage to the\n"
				"device. This is not recommended."));
	return initing;
}

// Set the progress bar fraction
void App::SetProgress(const gdouble fraction)
{
	m_Progress.set_fraction(std::min(1.0, fraction));
}

// Pulse the progress bar while retrieving the tracks from the device
int App::PulseProgress(const uint64_t received, const uint64_t total,
		const void* data)
{
	Playedit::App* app = (Playedit::App*)data;
	app->SetProgress((received * 1.0) / total);
	app->ProcessPending();
	return 0;
}

// Remove all references to a track from the playlist view.  Called when an
// track is deleted from the device.
void App::PlaylistClearTrack(uint32_t track_id)
{
	m_PlaylistView->ClearTrack(track_id);
}

// Initiate the device in its own thread
static void InitDevice(bool& initing, MTPDevice& device, int devnum)
{
	device.Connect(devnum);
	initing = false;
}

// Get device, get albums, tracks, and playlists.  Sort them and display them.
void App::ConnectDevice(int devnum)
{
	bool initing = true;
	static sigc::connection delete_event_id;

	// Don't allow the user to try to connect, quit, or disconnect while
	// connecting
	if (delete_event_id) delete_event_id.disconnect();
	delete_event_id = signal_delete_event().connect(
			sigc::bind<bool&>(&ForceQuit, initing));
	m_UIManager->get_action("/MainMenu/File/Quit")->set_sensitive(false);
	m_DeviceInteractionGroup->set_sensitive(false);

	m_Progress.set_pulse_step(0.1);

	// Initiate the device in a thread so we can pulsate the progress bar
	SetStatus(_("Initializing MTP device..."));
	Glib::Threads::Thread *thread = Glib::Threads::Thread::create(
			sigc::bind<bool&, MTPDevice&, int>(&InitDevice, initing, m_Device, devnum));
	while (initing)
	{
		m_Progress.pulse();
		ProcessPending();
		Glib::usleep(60000);
	}
	thread->join();

	// Get the data if we got the device
	if (m_Device)
	{
		// Get albums and tracks and put them into the music store
		SetStatus(_("MTP Device connected.  Getting tracks and albums..."));
		std::vector<album_pair_t> album_list;
		m_Device.GetMusicLists(album_list);
		m_MusicView->PopulateStore(album_list);

		// Get playlists and put them into the playlist store
		SetStatus(_("MTP Device connected.  Getting playlists..."));
		std::vector< Glib::RefPtr<playlist_t> > playlist_list;
		m_Device.GetPlaylistList(playlist_list);
		m_PlaylistView->PopulateStore(playlist_list, m_MusicView);

		// Clear the status bar
		UnsetStatus();

		// Enable device actions
		m_DeviceActionGroup->set_sensitive();
		m_DeviceActionGroup->set_visible();

		// Make the device interactions sensitive
		m_DeviceInteractionGroup->set_sensitive();

		// Set the application title to display the device name
		set_title("PlayEdit - " + m_Device.DeviceName());

		m_Device.SetReady();
	}
	else
		SetStatus(_("No MTP devices found, or unable to connect to MTP device."));

	m_Progress.set_fraction(0.0);

	// Let the user quit again
	if (delete_event_id) delete_event_id.disconnect();
	m_UIManager->get_action("/MainMenu/File/Quit")->set_sensitive(true);
}

void App::DisconnectDevice()
{
	// Disable device actions
	m_DeviceActionGroup->set_sensitive(false);
	m_DeviceActionGroup->set_visible(false);

	// Make the device interactions insensitive
	m_DeviceInteractionGroup->set_sensitive(false);

	// Release the device
	m_Device.Disconnect();

	// Clear the tree stores
	Glib::RefPtr<Gtk::TreeStore>::cast_static(m_PlaylistView->get_model())->clear();
	Glib::RefPtr<Gtk::TreeStore>::cast_static(m_MusicView->get_model())->clear();

	// Reset the application title
	set_title("PlayEdit");
}

void App::RenameDevice()
{
	EntryDialog dialog("New Device Name", "Enter new name for device",
			m_Device ? m_Device.DeviceName() : "");
	if (dialog.run() == Gtk::RESPONSE_OK)
	{
		m_Device.SetDeviceName(dialog.GetEntry());
		set_title("PlayEdit - " + m_Device.DeviceName());
	}
}

}
