// Copyright (C) 2009 Playedit.

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "tooltip.h"

#include <glibmm/i18n.h>
#include <glibmm/regex.h>
#include <cmath>
#include <iomanip>

// Returns a new string with "&"s replaced with "&amp;".
static Glib::ustring EscapeAmpersands(const Glib::ustring& data)
{
	Glib::RefPtr<Glib::Regex> regex = Glib::Regex::create("&");
	return regex->replace(data, 0, "&amp;", static_cast<Glib::RegexMatchFlags>(0));
}

// Adds marked up metadata to the end of text.
static void AddMetaData(Glib::ustring& text,
		const Glib::ustring& header, const Glib::ustring& data)
{
	text += (text.empty() ? "<b>" : "\n<b>") + header + ":</b> " + EscapeAmpersands(data);
}

// Set the tooltip text for a track.
Glib::ustring GetTrackTooltipText(const Glib::RefPtr<const track_t>& track)
{
	Glib::ustring text;

	AddMetaData(text, _("Title"), track->Name());
	if (track->RawAlbum().size())
		AddMetaData(text, _("Album"), track->RawAlbum());
	if (track->RawArtist().size())
		AddMetaData(text, _("Artist"), track->RawArtist());
	if (track->Composer().size())
		AddMetaData(text, _("Composer"), track->Composer());
	if (track->Genre().size())
		AddMetaData(text, _("Genre"), track->Genre());
	if (track->Date().size())
		AddMetaData(text, _("Date"), track->Date());
	if (track->Duration())
		AddMetaData(text, _("Duration"), Glib::ustring::format(
					track->Duration() / 60000, ":",
					std::setfill(L'0'), std::setw(2), (track->Duration() % 60000) / 1000));
	if (track->Rating())
		AddMetaData(text, _("Rating"), Glib::ustring::format(track->Rating()));
	if (track->UseCount())
		AddMetaData(text, _("Number of Times Played"),
				Glib::ustring::format(track->UseCount()));

	return text;
}

// Set the tooltip text for an album.
Glib::ustring GetAlbumTooltipText(const Glib::RefPtr<const album_t>& album)
{
	Glib::ustring text;

	AddMetaData(text, _("Album"), album->Name());
	AddMetaData(text, _("Artist"), album->Artist());
	if (album->Genre().size())
		AddMetaData(text, _("Genre"), album->Genre());
	if (album->NumTracks())
		AddMetaData(text, _("Number of Tracks"),
				Glib::ustring::format(album->NumTracks()));

	return text;
}

// Set the tooltip text for a playlist
Glib::ustring GetPlaylistTooltipText(const Glib::RefPtr<const playlist_t>& playlist,
		unsigned int duration)
{
	Glib::ustring text;

	AddMetaData(text, _("Playlist"), playlist->Name());
	AddMetaData(text, _("Number of Tracks"),
			Glib::ustring::format(playlist->NumTracks()));

	Glib::ustring minutes =
		Glib::ustring::format(std::setfill(L'0'), std::setw(2), (duration % 3600000) / 60000);
	Glib::ustring seconds =
		Glib::ustring::format(std::setfill(L'0'), std::setw(2), (duration % 60000) / 1000);
	if (duration)
		AddMetaData(text, _("Duration"), Glib::ustring::format(duration / 3600000, ":",
					minutes, ":", seconds));

	return text;
}
